'use strict';

var promise = require('promise');

var queryResult = {
    one: 1,
    many: 2,
    none: 4
};

exports.PGTransactPromise = function (pg, cfg) {

    var $p = function (func) {
        return new promise(func);
    };

    var self = this;

    // IMPORTANT: The caller must invoke done() when finished.
    this.connect = function () {
        return $p(function (resolve, reject) {
            pg.connect(cfg, function (err, client, done) {
                if (err) {
                    reject(err);
                } else {
                    resolve({
                        client: client,
                        done: done
                    });
                }
            });
        });
    };

    // Simple query, with opening and closing connection;
    this.query = function (query, qt) {
        return $p(function (resolve, reject) {
            self.connect()
                .then(function (db) {
                    self.monitor(true);
                    self.tQuery(db.client, query, qt)
                        .then(function (data) {
                            db.done();
                            resolve(data);
                            self.monitor(false);
                        }, function (reason) {
                            db.done();
                            reject(reason);
                            self.monitor(false);
                        });
                }, function (reason) {
                    reject(reason); // connection failed;
                });
        });
    };

    // query within another connection or transaction;
    this.tQuery = function (client, query, qt) {
        return $p(function (resolve, reject) {
            client.query(query, function (err, result) {
                if (err) {
                    reject(err.message);
                } else {
                    var data = result.rows;
                    var l = result.rows.length;
                    if (l) {
                        if (l > 1 && !(qt & queryResult.many)) {
                            reject('Single record was expected from query: ' + query);
                        } else {
                            if (!(qt & queryResult.many)) {
                                data = result.rows[0];
                            }
                        }
                    } else {
                        if (qt & queryResult.none) {
                            data = null;
                        } else {
                            reject('No records returned from query: ' + query);
                        }
                    }
                    resolve(data);
                }
            });
        });
    };

    // transaction block;
    this.transact = function (promiseFunc) {
        var db, t_data, t_reason, success;
        return $p(function (resolve, reject) {
            self.connect()
                .then(function (data) {
                    self.monitor(true);
                    db = data;
                    return self.tQuery(db.client, 'begin', queryResult.none);
                }, function (reason) {
                    reject(reason); // connection issue;
                })
                .then(function () {
                    promiseFunc(db.client)
                        .then(function (data) {
                            t_data = data;
                            success = true;
                            return self.tQuery(db.client, 'commit', queryResult.none);
                        }, function (reason) {
                            t_reason = reason;
                            success = false;
                            return self.tQuery(db.client, 'rollback', queryResult.none);
                        })
                        .then(function () {
                            db.done();
                            self.monitor(false);
                            // either commit or rollback successfully executed;
                            if (success) {
                                resolve(t_data);
                            } else {
                                reject(t_reason);
                            }
                        }, function (reason) {
                            // either commit or rollback failed;
                            db.done();
                            reject(reason);
                            self.monitor(false);
                        });
                }, function (reason) {
                    db.done();
                    reject(reason); // issue with 'begin' command;
                    self.monitor(false);
                });
        });
    };

    this.monitor = function (open) {
        if (open) {
            self.cnCounter++;
        } else {
            self.cnCounter--;
        }
    };

    this.cnCounter = 0; // internal connection counter;

    this.none = function (query) {
        return self.query(query, queryResult.none);
    };

    this.one = function (query) {
        return self.query(query, queryResult.one);
    };

    this.many = function (query) {
        return self.query(query, queryResult.many);
    };

    this.oneOrNone = function (query) {
        return self.query(query, queryResult.one | queryResult.none);
    };

    this.manyOrNone = function (query) {
        return self.query(query, queryResult.many | queryResult.none);
    };

    this.as = {
        bool: function (val) {
            if (typeof(val) === 'undefined' || val === null) {
                return 'null';
            }
            return val ? 'TRUE' : 'FALSE';
        },
        // TODO: must replace quota symbols here;
        text: function (val) {
            if (typeof(val) === 'undefined' || val === null) {
                return 'null';
            }
            return "'" + val + "'";
        },
        date: function (val) {
            if (typeof(val) === 'undefined' || val === null) {
                return 'null';
            }
            if (val instanceof Date) {
                return "'" + val.toUTCString() + "'";
            } else {
                throw new Error("'" + val + "' doesn't represent a valid Date object or value");
            }
        }
    };
};

/*
var $db = {
    /////////////////////////////////
    // base methods;
    /////////////////////////////////
    // specific methods;
    func: function (funcName, params) {
        return this.one(this.createFuncQuery(funcName, params));
    },
    proc: function (procName, params) {
        return this.oneOrNull(this.createFuncQuery(procName, params));
    },
    // wraps up one value and returns it ready to be inserted into the database;
    wrap: function (val) {
        if (!val) {
            return 'null';
        }
        switch (typeof(val)) {
            case 'string':
                return "'" + val + "'";
            case 'boolean':
                return val ? 'TRUE' : 'FALSE';
            default:
                if (val instanceof Date) {
                    return "'" + val.toUTCString() + "'";
                } else {
                    return val;
                }
        }
    },
    formatValues: function (values) {
        var s = '';
        if (Array.isArray(values) && values.length > 0) {
            for (var i = 0; i < values.length; i++) {
                if (i > 0) {
                    s += ',';
                }
                s += this.wrap(values[i]);
            }
        }
        return s;
    },
    createFuncQuery: function (funcName, params) {
        return 'select * from ' + funcName + '(' + this.formatValues(params) + ');';
    }

}
*/
